package biz.mbisoftware.fn.ejb.session.common;

import java.util.List;
import java.util.Locale;

import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.TypedQuery;

import org.apache.log4j.Logger;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import biz.mbisoftware.common.StringHelper;
import biz.mbisoftware.common.XmlDocument;
import biz.mbisoftware.common.exceptions.MbiException;
import biz.mbisoftware.fn.ejb.entity.MbiSynon;
import biz.mbisoftware.fn.ejb.entity.MbiSyueb;
import biz.mbisoftware.fn.ejb.session.QueryFactory;
import biz.mbisoftware.fn.ejb.session.VersionFacade;
import biz.mbisoftware.fn.wstypes.GenericCharacterValue;
import biz.mbisoftware.fn.wstypes.GenericIntegerReference;
import biz.mbisoftware.fn.wstypes.ServiceByteStreamResponse;
import biz.mbisoftware.fn.wstypes.ServiceResponse;
import biz.mbisoftware.fn.wstypes.ServiceResponse.ErrorCode;
import biz.mbisoftware.fn.wstypes.ServiceStringResponse;

/**
 * Translation of synonyms.
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
public class MbiSyuebFacade
{
    /** LOGGER to log errors. */
    private static final Logger LOGGER = Logger.getLogger( MbiSyuebFacade.class );

    /** The translated synonym. */
    private String globTranslation;

    /** Select statement to read data from mbi_syueb. */
    private static final String QUERY =
            "SELECT mbiSyueb.uebersetzung FROM MbiSynon AS mbiSynon"
            + " JOIN mbiSynon.mbiSyueb AS mbiSyueb"
            + " WHERe mbiSynon.synonName  = :synonName"
            + " AND mbiSyueb.obClass = 0 AND mbiSyueb.msName = ' '"
            + " AND mbiSyueb.projektKnz in ( ' ', :projektKnz )"
            + " AND mbiSyueb.sprachKbez = :sprachKbez"
            + " ORDER BY mbiSyueb.projektKnz DESC";
    
    /** The class's entity manager. */
    @PersistenceContext
    private EntityManager entityManager;
    
    /**
     * Translate a synonym in the specified language.
     * @param synonymLiteral the synonym to find a translation for
     * @param language language
     * @return if a translation was found
     * @throws MbiException if an error occurs
     */
    private boolean getTranslation( final String synonymLiteral, final String language )
    {
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( synonymLiteral + ", " + language );
        }
        int numFound = 0;
        boolean endsWithColon = false;
        String[] synonyms = synonymLiteral.split( "[~\\s]" );
        StringBuilder resultSb = new StringBuilder();
        for ( int i = 0; i < synonyms.length; i++ ) {
            String synonym = synonyms[i].trim();
            boolean isCite = false;
            if ( i == synonyms.length - 1 && synonym.endsWith( ":" ) ) {
                endsWithColon = true;
                synonym = synonym.substring( 0, synonym.length() - 1 );
            } else if ( synonym.startsWith( "'" ) && synonym.endsWith( "'" ) ) {
                synonym = synonym.substring( 1, synonym.length() - 1 );
                isCite = true;
            }
            String projectShortName;
            try {
                projectShortName = VersionFacade.getProjectShortName();
            } catch ( MbiException e ){
                projectShortName = " ";
            }
            if ( synonym.matches( "[A-Z0-9_]+" ) ) {
                TypedQuery<String> query = QueryFactory.createQuery( this.entityManager, MbiSyuebFacade.QUERY, String.class, true, true, true );
                query.setParameter( "synonName", StringHelper.fillString( synonym, MbiSynon.LEN_SYNON_NAME ) );
                query.setParameter( "sprachKbez", StringHelper.fillString( language, MbiSyueb.LEN_SPRACH_KBEZ ) );
                query.setParameter( "projektKnz", projectShortName );
                List<String> translations = query.getResultList();
                if ( !translations.isEmpty() ) {
                    if ( resultSb.length() > 0 ) {
                        resultSb.append( ' ' );
                    }
                    String uebersetzung = translations.get( 0 );
                    if ( uebersetzung != null ) {
                        if ( isCite ) {
                            resultSb.append( "'" );
                        }
                        resultSb.append( uebersetzung );
                        if ( isCite ) {
                            resultSb.append( "'" );
                        }
                    }
                    numFound++;
                }
            } else {
                if ( resultSb.length() > 0 ) {
                    resultSb.append( ' ' );
                }
                resultSb.append( synonym );
                numFound++;
            }
        }
        boolean found;
        if ( numFound == synonyms.length ) {
            found = true;
            if ( endsWithColon ) {
                resultSb.append( ':' );
            }
            this.globTranslation = resultSb.toString();
        } else {
            found = false;
            this.globTranslation = null;
        }
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( Boolean.valueOf( found ) );
        }
        return found;
    }

    public ServiceStringResponse getTranslation( final String synonym, final Locale[] locales )
    {
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( synonym + ", " + locales );
        }
        String translation = null;
        ErrorCode errorCode = ServiceResponse.ErrorCode.ERROR;
        if ( synonym == null ) {
            errorCode = ServiceResponse.ErrorCode.SUCCESS;
        } else {
            if ( locales != null ) {
                for ( int iIdx = 0; iIdx < locales.length; iIdx++ ) {
                    if ( this.getTranslation( synonym, locales[iIdx].getLanguage() ) ) {
                        errorCode = ServiceResponse.ErrorCode.SUCCESS;
                        translation = this.globTranslation;
                        break;
                    }
                }
            }
            if ( translation == null ) {
                if ( this.getTranslation( synonym, "de" ) ) {
                    errorCode = ServiceResponse.ErrorCode.WARNING;
                    translation = this.globTranslation;
                }
            }
            if ( translation == null ) {
                errorCode = ServiceResponse.ErrorCode.ERROR;
                translation = synonym;
            }
        }
        ServiceStringResponse result = new ServiceStringResponse();
        result.setErrorCode( errorCode );
        result.setValue( translation );
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( result );
        }
        return result;
    }

    public GenericIntegerReference[] translate( final GenericIntegerReference[] originalParts, final Locale[] locales )
    {
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( originalParts + ", " + locales );
        }
        GenericIntegerReference[] parts = new GenericIntegerReference[originalParts.length];
        for ( int iIdx = 0; iIdx < originalParts.length; iIdx++ ) {
            parts[iIdx] = new GenericIntegerReference();
            parts[iIdx].setId( originalParts[iIdx].getId() );
            parts[iIdx].setDescription( this.getTranslation( originalParts[iIdx].getDescription(), locales ).getValue() );
        }
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( parts );
        }
        return parts;
    }

    public GenericCharacterValue[] translate( final GenericCharacterValue[] originalParts, final Locale[] locales )
    {
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( originalParts + ", " + locales );
        }
        GenericCharacterValue[] parts = new GenericCharacterValue[originalParts.length];
        for ( int iIdx = 0; iIdx < parts.length; iIdx++ ) {
            parts[iIdx] = new GenericCharacterValue();
            parts[iIdx].setValue( originalParts[iIdx].getValue() );
            parts[iIdx].setDescription( this.getTranslation( originalParts[iIdx].getDescription(), locales ).getValue() );
        }
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( parts );
        }
        return parts;
    }

    /**
     * Translate a xml document tree in one of the languages defined by locales.
     * @param xmlDoc The xml document tree to translate
     * @param locales list of locales
     * @return the translation
     */
    private ServiceStringResponse translate( final XmlDocument xmlDoc, final Locale[] locales )
    {
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( xmlDoc + ", " + locales );
        }
        ServiceStringResponse response = new ServiceStringResponse();
        Element rootNode = xmlDoc.getDocumentElement();
        try {
            this.translateElement( rootNode, locales );
            response.setErrorCode( ServiceResponse.ErrorCode.SUCCESS );
            response.setValue( xmlDoc.transformDOM2String() );
        } catch ( MbiException e ) {
            response.setErrorMessagesFromException( e );
        }
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( response );
        }
        return response;
    }

    /**
     * Translate a xml document tree in one of the languages defined by locales.
     * @param xmlDoc The xml document tree to translate
     * @param locales list of locales
     * @return the translation
     */
    private ServiceByteStreamResponse translateToByteStream( final XmlDocument xmlDoc, final Locale[] locales )
    {
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( xmlDoc + ", " + locales );
        }
        ServiceByteStreamResponse response = new ServiceByteStreamResponse();
        Element rootNode = xmlDoc.getDocumentElement();
        try {
            this.translateElement( rootNode, locales );
            response.setErrorCode( ServiceResponse.ErrorCode.SUCCESS );
            response.setValue( xmlDoc.transformDOM2ByteStream() );
        } catch ( MbiException e ) {
            response.setErrorMessagesFromException( e );
        }
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( response );
        }
        return response;
    }

    public ServiceStringResponse translateFile( final String filename, final Locale[] locales ) throws MbiException
    {
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( filename + ", " + locales );
        }
        XmlDocument xmlDoc = new XmlDocument( XmlDocument.FILE, "META-INF/" + filename );
        ServiceStringResponse result = this.translate( xmlDoc, locales );
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( result );
        }
        return result;
    }

    public ServiceByteStreamResponse translateXmlStream( final byte[] xmlStream, final Locale[] locales ) throws MbiException
    {
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( xmlStream + ", " + locales );
        }
        XmlDocument xmlDoc = new XmlDocument( xmlStream );
        ServiceByteStreamResponse result = this.translateToByteStream( xmlDoc, locales );
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( result );
        }
        return result;
    }

    /**
     * Translate the TextNode's content or go one level deeper.
     * @param node The current node.
     * @param locales The language codes needed for translation.
     * @throws MbiException if an error occures
     */
    private void translateElement( final Node node, final Locale[] locales )
    {
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( node + ", " + locales );
        }
        NodeList nl = node.getChildNodes();
        for ( int i = 0; i < nl.getLength(); i++ ) {
            Node child = nl.item( i );
            if ( child.getNodeType() == Node.ELEMENT_NODE ) {
                this.translateElement( child, locales );
            } else if ( child.getNodeType() == Node.TEXT_NODE ) {
                String s = child.getNodeValue();
                if ( s != null && s.trim().length() > 1 && s.matches( "[A-Z0-9_~]+" ) ) {
                    ServiceStringResponse ssr = this.getTranslation( s, locales );
                    child.setNodeValue( ssr.getValue() );
                }
            }
        }
        if ( MbiSyuebFacade.LOGGER.isDebugEnabled() ) {
            MbiSyuebFacade.LOGGER.debug( "" );
        }
    }
}
